from re import compile, escape, Pattern
from .mreplace import MultipleReplace
from .tokens import ToTokens
from typing import Dict, List, Optional

_groups: Dict[str, List[str]] = {
    "[benzene]": ["cccccc6"],
    "[Ts]": ["S!!:oocccccc6..C...", "S!!:occcccc6..C...o", "S!!:cccccc6..C...oo"],
    "[Tf]": ["S!!:ooCFFF", "S!!:oCFFFo", "S!!:CFFFoo"],
    "[Ms]": ["S!!:ooC.", "S!!:oC.o", "S!!:C.oo"],
    "[Piv]": ["coCC.C.C.", "cCC.C.C.o"],
    "[Boc]": ["coOCC.C.C.", "cOCC.C.C.o"],
    "[Tol]": ["cccccc6..C..."],
    "[Cbz]": ["coOCcccccc6......", "cOCcccccc6......o"],
    "[Bn]": ["Ccccccc6......"],
    "[Bz]": ["cocccccc6....."],
    "[Ph]": ["cccccc6....."],
    "[OEt]": ["OCC.."],
    "[OMe]": ["OC."],
    "[NHAc]": ["NcoC..", "NcC.o."],
    "[NHMe]": ["NC.."],
    "[NMe2]": ["NC.C."],
    "[OAc]": ["OcoC.", "OcC.o"],
    "[COOEt]": ["coOCC..", "cOoCC.."],
    "[COOMe]": ["coOC.", "cOoC."],
    "[COO-]": ["coO-", "cO-o"],
    "[NO2]": ["n+oO-", "n+O-o"],
    "[Ac]": ["coC.", "cC.o"],
    "[COOH]": ["coO.", "cO.o"],
    "[CHO]": ["co."],
    "[tBu]": ["CC.C.C."],
    "[nBu]": ["CCCC...."],
    "[sBu]": ["CC.CC...", "CCC..C.."],
    "[iBu]": ["CCC.C..."],
    "[nPr]": ["CCC..."],
    "[iPr]": ["CC.C."],
    "[Et]": ["CC.."],
    "[CN]": ["C:N:"],
    "[CF3]": ["CFFF"],
    "[CCl3]": ["C[Cl][Cl][Cl]"],
}


def Groups() -> Dict[str, List[str]]:
    """Keys are functional group abbreviations, values are lists of one or more AMSR strings
    consisting only of atom/bond tokens.  May be modified, but :func:`InitializeGroups`
    must be called after modification.

    :return: Groups dictionary
    """
    return _groups


_mr: Optional[MultipleReplace] = None
_pattern: Optional[Pattern] = None


def InitializeGroups() -> None:
    """Initialize tree and compile regular expression for converting between
    group abbreviations and tokens.  Must be called after modification of
    :func:`Groups` dictionary.
    """
    global _pattern, _mr
    _mr = MultipleReplace([(ToTokens(g), k) for k, v in _groups.items() for g in v])
    _pattern = compile("(" + "|".join([escape(k) for k in _groups.keys()]) + ")")


InitializeGroups()


def DecodeGroups(s: str) -> str:
    return s if _pattern is None else _pattern.sub(lambda m: _groups[m.group(1)][0], s)


def EncodeGroups(s: List[str]) -> List[str]:
    return s if _mr is None else _mr.replace(s)



from rdkit import Chem
from networkx import Graph
from networkx.algorithms import max_weight_matching
from .ringinfo import BridgeheadAtoms, RingAtoms


class PiBonds:
    def piBond(self, i, j, order):
        self.mol.GetBondBetweenAtoms(i, j).SetBondType(
            Chem.BondType.TRIPLE if order == 3 else Chem.BondType.DOUBLE
        )
        self.atom[i].nPiBonds += order - 1
        self.atom[j].nPiBonds += order - 1

    def incrementBond(self, i, j):
        b = self.mol.GetBondBetweenAtoms(i, j)
        t = b.GetBondType()
        if t == Chem.BondType.SINGLE:
            b.SetBondType(Chem.BondType.DOUBLE)
        elif t == Chem.BondType.DOUBLE:
            b.SetBondType(Chem.BondType.TRIPLE)
        self.atom[i].nPiBonds += 1
        self.atom[j].nPiBonds += 1

    def canPiBond(self, i):
        return self.atom[i].nAvailablePiBonds() >= 1 and i not in self.excluded

    def canMultiplePiBond(self, i):
        return self.atom[i].nAvailablePiBonds() >= 2 and i not in self.excluded

    def canBePi(self, i, j):
        return self.canPiBond(i) and self.canPiBond(j)

    def canBeTriple(self, i, j):
        return self.canMultiplePiBond(i) and self.canMultiplePiBond(j)

    def possiblePiBonds(self):
        for b in self.mol.GetBonds():
            i, j = b.GetBeginAtomIdx(), b.GetEndAtomIdx()
            if self.canBePi(i, j):
                yield i, j

    def isCarbon(self, i):
        return self.atom[i].isCarbon()

    def reduceGraph(self):
        self.graph = self.graph.subgraph(
            [i for i in self.graph.nodes if self.canPiBond(i)]
        )

    def singleCoordinate(self, i):
        return self.graph.degree[i] == 1

    def __init__(self, mol, atom, useFilters=True):

        self.mol = mol
        self.atom = atom
        self.excluded = set()

        if useFilters:
            # use a subset of filters from GDB17
            # Ruddigkeit et al., J. Chem. Inf. Model. 52, 2864 (2012)
            Chem.GetSSSR(self.mol)
            ringInfo = self.mol.GetRingInfo()
            self.excluded.update(BridgeheadAtoms(mol, ringInfo, 7))
            self.excluded.update(RingAtoms(ringInfo, 3))

        # subgraph of atoms that can make pi bonds
        self.graph = Graph(self.possiblePiBonds())

        # single coordinate - heteroatoms first
        done = False
        while not done:
            done = True
            for i in sorted(self.graph.nodes, key=lambda k: self.isCarbon(k)):
                if self.singleCoordinate(i):
                    for j in self.graph.neighbors(i):
                        if self.canBeTriple(i, j):
                            self.piBond(i, j, 3)
                        elif self.canBePi(i, j):
                            self.piBond(i, j, 2)
                        else:
                            continue
                        done = False
            if not done:
                self.reduceGraph()

        # isolated sp centers
        for i in self.graph.nodes:
            if self.canMultiplePiBond(i):
                for j in self.graph.neighbors(i):
                    if self.canMultiplePiBond(j):
                        break
                else:
                    for j in self.graph.neighbors(i):
                        if self.canBePi(i, j):
                            self.piBond(i, j, 2)

        # remaining pi bonds
        done = False
        while not done:
            done = True
            self.reduceGraph()
            for i, j in max_weight_matching(self.graph):
                self.incrementBond(i, j)
                done = False



class MultipleReplace:
    def __init__(self, a):
        self.root = ({}, {})
        for k, v in a:
            n, val = self.root
            for i in k[:-1]:
                if i not in n:
                    n[i] = ({}, {})
                n, val = n[i]
            val[k[-1]] = v

    def __repr__(self):
        return self.root.__repr__()

    def find(self, k, nk, i):
        n, val = self.root
        target_val = None
        target_depth = None
        depth = 1
        while i < nk:
            if k[i] in val:
                target_val = val[k[i]]
                target_depth = depth
            if k[i] in n:
                n, val = n[k[i]]
                depth += 1
            else:
                break
            i += 1
        return target_val, target_depth

    def replace(self, s):
        t = []
        i = 0
        n = len(s)
        while i < n:
            val, depth = self.find(s, n, i)
            if val is not None:
                t.append(val)
                i += depth
            else:
                t.append(s[i])
                i += 1
        return t



from rdkit import Chem
from .encode import FromMol
from .decode import ToMol


def CheckMol(m: Chem.Mol, useFilters: bool = False) -> bool:
    """Do round trip from RDKit mol m to AMSR.
    Compare InChI strings (with -FixedH) before and after round trip.

    :param m: RDKit Mol
    :param useFilters: use a subset of filters from J. Chem. Inf. Model. 52, 2864 (2012) to exclude unstable or synthetically inaccessible molecules
    :return: do InChI strings match?
    """
    i1 = Chem.MolToInchi(m, options="-FixedH")
    i2 = Chem.MolToInchi(ToMol(FromMol(m), useFilters=useFilters), options="-FixedH")
    if i1 == i2:
        return True
    else:
        print(i1)
        print(i2)
        return False


def CheckSmiles(s: str, useFilters: bool = False) -> bool:
    """Convert SMILES s to RDKit Mol, then do round trip to AMSR.
    Compare InChI strings (with -FixedH) before and after round trip.

    :param s: SMILES
    :param useFilters: use a subset of filters from J. Chem. Inf. Model. 52, 2864 (2012) to exclude unstable or synthetically inaccessible molecules
    :return: do InChI strings match?
    """
    return CheckMol(Chem.MolFromSmiles(s), useFilters=useFilters)



from rdkit import Chem
from re import match, sub
from .valence import VALENCE, BANGS
from .parity import IsEvenParity
from .tokens import CW, CCW, PLUS, MINUS, RADICAL, EXTRA_PI, BANG


def GetSeenIndex(a):
    return a.GetIntProp("_seenIndex")


def SetSeenIndex(a, i):
    return a.SetIntProp("_seenIndex", i)


def IsSeen(a):
    return a.HasProp("_seenIndex")


class Atom:
    def __init__(self, sym):
        self.sym = sym
        self.bangs = sym.count(BANG)
        m = match(r"\[(\d+)", sym)
        self.isotope = None if m is None else int(m.group(1))
        self.chg = sym.count(PLUS) - sym.count(MINUS)
        self.nrad = sym.count(RADICAL)
        if CCW in sym:
            self.ct = Chem.ChiralType.CHI_TETRAHEDRAL_CCW
        elif CW in sym:
            self.ct = Chem.ChiralType.CHI_TETRAHEDRAL_CW
        else:
            self.ct = Chem.ChiralType.CHI_UNSPECIFIED
        self.atomSym = sub(r"[^A-Za-z]", "", sym)
        if self.atomSym[0].islower():
            self.atomSym = self.atomSym[0].upper() + self.atomSym[1:]
            self.maxPiBonds = 1
        else:
            self.maxPiBonds = 0
        self.maxPiBonds += 2 * sym.count(EXTRA_PI)
        self.nPiBonds = 0
        self.maxNeighbors = (
            VALENCE[(self.atomSym, self.chg, self.bangs)] - self.nrad - self.maxPiBonds
        )
        self.nNeighbors = 0
        self.isSaturated = False

    def canBond(self):
        return (not self.isSaturated) and self.nNeighbors < self.maxNeighbors

    def nAvailablePiBonds(self):
        return self.maxPiBonds - self.nPiBonds

    def asRDAtom(self):
        a = Chem.Atom(self.atomSym)
        a.SetFormalCharge(self.chg)
        a.SetNumRadicalElectrons(self.nrad)
        a.SetChiralTag(self.ct)
        if self.isotope:
            a.SetIsotope(self.isotope)
        return a

    def isCarbon(self):
        return self.atomSym == "C"

    def symWith(self, s):
        sym = self.sym
        if sym.startswith("["):
            return "[" + sym[1:-1] + s + "]"
        else:
            return sym + s

    def asToken(self, a, mol):
        ct = a.GetChiralTag()
        isEven = IsEvenParity([GetSeenIndex(b) for b in a.GetNeighbors()])
        if ct == Chem.ChiralType.CHI_TETRAHEDRAL_CCW:
            return self.symWith(CCW if isEven else CW)
        elif ct == Chem.ChiralType.CHI_TETRAHEDRAL_CW:
            return self.symWith(CW if isEven else CCW)
        else:
            return self.sym

    @classmethod
    def fromRDAtom(cls, a):
        atomSym = a.GetSymbol()
        chg = a.GetFormalCharge()
        valence = a.GetTotalValence()
        nrad = a.GetNumRadicalElectrons()
        isotope = a.GetIsotope()
        if atomSym == "He" or nrad > 4:  # RDKit weirdness?
            nrad = 0
        bangs = BANGS.get((atomSym, chg, valence), 0)
        q, r = divmod(VALENCE[(atomSym, chg, bangs)] - nrad - a.GetTotalDegree(), 2)
        c = f"{PLUS*chg if chg > 0 else ''}{MINUS*(-chg) if chg < 0 else ''}{RADICAL*nrad}{BANG*bangs}{EXTRA_PI*q}"
        sym = (f"{isotope}" if isotope else "") + (atomSym.lower() if r else atomSym)
        return cls(f"[{sym}{c}]" if len(atomSym) == 2 or isotope else f"{sym}{c}")



from itertools import combinations
from .count import Count


def BridgeheadAtoms(mol, ringInfo, n):
    for i, j in combinations([r for r in ringInfo.BondRings() if len(r) <= n], 2):
        bridge = set(i) & set(j)
        if len(bridge) >= 2:
            a = {}
            for b in bridge:
                b = mol.GetBondWithIdx(b)
                Count(a, b.GetBeginAtomIdx())
                Count(a, b.GetEndAtomIdx())
            yield from (k for k, n in a.items() if n == 1)


def RingAtoms(ringInfo, n):
    yield from (i for r in ringInfo.AtomRings() if len(r) == n for i in r)



from rdkit import Chem
from typing import Optional, List, Set, FrozenSet
from .atom import Atom, GetSeenIndex, SetSeenIndex, IsSeen
from .bfs import BFSFind
from .bond import Bond
from .groups import EncodeGroups
from .tokens import DOT, SKIP


def _ringTokens(n, nSkip):
    yield f"[{n}]" if n > 9 else f"{n}"
    yield from iter(SKIP * nSkip)


def _searchOrder(b, a):
    # 1. seen atoms before unseen atoms (i.e. rings)
    # 2. aromatic bonds first
    # 3. small rings before larger (for seen) .. otherwise atom index (for unseen)
    c = b.GetOtherAtom(a)
    isSeen = IsSeen(c)
    return (
        not isSeen,
        not b.GetIsAromatic(),
        GetSeenIndex(a) - GetSeenIndex(c) if isSeen else c.GetIdx(),
    )


def _isDot(t):
    return t == DOT


def _removeTrailingDots(t):
    i = len(t) - 1
    while i >= 0:
        if not _isDot(t[i]):
            del t[i + 1 :]
            break
        i -= 1
    return t


def _bondTokens(b, bond):
    if bond is not None:
        yield (b, bond)


def FromMolToTokens(mol: Chem.Mol, useGroups: Optional[bool] = True) -> List[str]:
    """Convert RDKit Mol to list of AMSR tokens

    :param mol: RDKit Mol
    :param useGroups: use group symbols/abbreviations
    :return: list of AMSR tokens
    """

    atom = [Atom.fromRDAtom(a) for a in mol.GetAtoms()]
    seenBonds: Set[FrozenSet[int]] = set()
    nSeenAtoms = 0

    def _search(a):
        nonlocal nSeenAtoms
        i = a.GetIdx()
        ai = atom[i]
        SetSeenIndex(a, nSeenAtoms)
        nSeenAtoms += 1
        for b in sorted(a.GetBonds(), key=lambda b: _searchOrder(b, a)):
            c = b.GetOtherAtom(a)
            j = c.GetIdx()
            ij = frozenset([i, j])
            if ij in seenBonds:
                continue
            aj = atom[j]
            bond = Bond.fromRD(b)
            if IsSeen(c):  # ring
                nSkip = 0
                for k, depth in BFSFind(a, j, seenBonds):
                    if k == j:
                        yield from _bondTokens(b, bond)
                        yield from _ringTokens(depth + 1, nSkip)
                        seenBonds.add(ij)
                        ai.nNeighbors += 1
                        aj.nNeighbors += 1
                        break
                    elif atom[k].canBond():
                        nSkip += 1
            else:  # new atom
                seenBonds.add(ij)
                ai.nNeighbors += 1
                aj.nNeighbors += 1
                yield from _bondTokens(b, bond)
                yield (c, aj)
                yield from _search(c)
        # end loop over bonds
        if ai.canBond():
            ai.isSaturated = True
            yield DOT

    def _getPreTokens():
        for i, a in enumerate(mol.GetAtoms()):
            if not IsSeen(a):
                yield a, atom[i]
                yield from _search(a)

    t = [
        t[1].asToken(t[0], mol) if type(t) == tuple else t
        for t in list(_getPreTokens())
    ]

    if useGroups:
        t = EncodeGroups(t)

    return _removeTrailingDots(t)


def FromMol(mol: Chem.Mol, useGroups: Optional[bool] = True) -> str:
    """Convert RDKit Mol to AMSR

    :param mol: RDKit Mol
    :param useGroups: use group symbols/abbreviations
    :return: AMSR
    """
    return "".join(FromMolToTokens(mol, useGroups=useGroups))


def FromSmiles(s: str, useGroups: Optional[bool] = True) -> str:
    """Convert SMILES to AMSR

    :param s: SMILES
    :param useGroups: use group symbols/abbreviations
    :return: AMSR
    """
    return FromMol(Chem.MolFromSmiles(s), useGroups=useGroups)


def FromSmilesToTokens(s, useGroups: Optional[bool] = True) -> List[str]:
    """Convert SMILES to list of AMSR tokens

    :param mol: RDKit Mol
    :param useGroups: use group symbols/abbreviations
    :return: AMSR
    """
    return FromMolToTokens(Chem.MolFromSmiles(s), useGroups=useGroups)



from Levenshtein import editops
from .encode import FromSmilesToTokens
from .decode import ToMol
from rdkit import Chem


class Morph:
    def __init__(self, s, t):
        s = list(s)[:]
        self.amsr = ["".join(s)]
        k = 0
        for op, i, j in editops(s, t):
            if op == "insert":
                s.insert(i + k, t[j])
                k += 1
            elif op == "delete":
                del s[i + k]
                k -= 1
            elif op == "replace":
                s[i + k] = t[j]
            self.amsr.append("".join(s))
        self.mol = []
        i0 = None
        for s in self.amsr:
            m = ToMol(s, contiguous=True, useFilters=True)
            i = Chem.MolToInchi(m, options="-FixedH")
            if i0 is None or i != i0:
                self.mol.append(m)
                i0 = i

    @classmethod
    def fromSmiles(cls, s, t):
        return cls(FromSmilesToTokens(s), FromSmilesToTokens(t))

    def showAsSmiles(self):
        for m in self.mol:
            print(Chem.MolToSmiles(m))



from rdkit import Chem
from .atom import GetSeenIndex
from .tokens import E, Z


class Bond:
    def __init__(self, sym=None):
        self.sym = sym

    def rdStereo(self):
        if self.sym == E:
            return Chem.BondStereo.STEREOE
        elif self.sym == Z:
            return Chem.BondStereo.STEREOZ
        else:
            return None

    def asToken(self, b, mol):
        a1 = b.GetBeginAtom()
        i1 = a1.GetIdx()
        a2 = b.GetEndAtom()
        i2 = a2.GetIdx()
        n1 = [GetSeenIndex(c) for c in a1.GetNeighbors() if c.GetIdx() != i2]
        n2 = [GetSeenIndex(c) for c in a2.GetNeighbors() if c.GetIdx() != i1]
        flip = False
        for a in b.GetStereoAtoms():
            i = GetSeenIndex(mol.GetAtomWithIdx(a))
            if i in n1 and i != min(n1):
                flip = not flip
            elif i in n2 and i != min(n2):
                flip = not flip
        if self.sym == E:
            return Z if flip else E
        if self.sym == Z:
            return E if flip else Z

    @classmethod
    def fromRD(cls, b):
        s = b.GetStereo()
        if s == Chem.BondStereo.STEREOE:
            return cls(E)
        if s == Chem.BondStereo.STEREOZ:
            return cls(Z)
        return None



from .encode import FromMolToTokens
from numpy import mean, std
from random import choices, gauss
from .count import Count


class Sampler:
    def __init__(self, mols):
        d = {}
        n = []
        for m in mols:
            tok = FromMolToTokens(m, useGroups=True)
            n.append(len(tok))
            for t in tok:
                Count(d, t)
        self.tokens = list(d.keys())
        self.weights = list(d.values())
        self.navg = mean(n)
        self.nstddev = std(n)

    def sampleTokens(self):
        return choices(
            population=self.tokens,
            weights=self.weights,
            k=max(round(gauss(self.navg, self.nstddev)), 1),
        )

    def sample(self):
        return "".join(self.sampleTokens())



import anytree
import rdkit.Chem as rd
from .atom import GetSeenIndex


def _idx(n):
    return n.name.GetIdx()


def BFSPath(n):
    if n is not None:
        yield _idx(n)
        yield from BFSPath(n.parent)


def BFSTree(a, n):
    ai = a.GetIdx()
    if n == 0:
        return [ai]
    root = anytree.Node(a)
    q = [root]
    seen = {ai}
    tree = []
    while q:
        node = q.pop(0)
        a = node.name
        for b in a.GetBonds():
            c = b.GetOtherAtom(a)
            ci = c.GetIdx()
            if ci not in seen:
                seen.add(ci)
                child = anytree.Node(c, parent=node)
                if child.depth < n:
                    q.append(child)
                elif ci < ai:
                    tree.append(child)
    return sorted(tree, key=_idx, reverse=True)


def BFSFind(a, targetIdx, seenBonds):
    ai = a.GetIdx()
    root = anytree.Node(a)
    q = [root]
    seen = {ai}
    targetNode = None
    while q:
        node = q.pop(0)
        a = node.name
        ai = a.GetIdx()
        for b in a.GetBonds():
            c = b.GetOtherAtom(a)
            ci = c.GetIdx()
            if frozenset([ai, ci]) in seenBonds and ci not in seen:
                seen.add(ci)
                child = anytree.Node(c, parent=node)
                if ci == targetIdx:
                    targetNode = child
                if targetNode is None:
                    q.append(child)
    return (
        (n.name.GetIdx(), n.depth)
        for n in sorted(
            anytree.LevelOrderIter(root, filter_=lambda n: n.depth == targetNode.depth),
            key=lambda n: GetSeenIndex(n.name),
            reverse=True,
        )
    )



import os
from glob import glob
from setuptools import setup

exec(open("amsr/version.py").read())

# with open("README.md", "r", encoding="utf-8") as fh:
#    long_description = fh.read()

long_description = "See [documentation](https://hstern2.github.io/amsr)"

setup(
    name="amsr",
    version=__version__,
    description="Another molecular string representation",
    author="Harry Stern",
    url="https://hstern2.github.io/amsr/",
    license="MIT",
    packages=["amsr"],
    install_requires=["rdkit-pypi", "networkx", "anytree", "Levenshtein"],
    test_suite="tests",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Chemistry",
    ],
)



from re import compile, escape, Match
from typing import Iterator, List

DOT = "."
E = "_"
Z = "^"
CW = ")"
CCW = "("
PLUS = "+"
MINUS = "-"
EXTRA_PI = ":"
BANG = "!"
RADICAL = "*"
L_BRACKET = "["
R_BRACKET = "]"
PERCENT = "%"
SKIP = ";"

_pbond = f"(?P<bond>[{escape(E)}{escape(Z)}])"
_c = f"[{escape(PLUS+MINUS+RADICAL+EXTRA_PI+BANG+CW+CCW)}]*"
_patom = (
    f"(?P<atom>{escape(L_BRACKET)}[0-9]*[A-Za-z]+{_c}{escape(R_BRACKET)}|[A-Za-z]{_c})"
)
_pring = (
    f"(?P<ring>({escape(L_BRACKET)}[0-9]+{escape(R_BRACKET)}|[3-9]){escape(SKIP)}*)"
)
_psaturate = f"(?P<saturate>{escape(DOT)})"

RegExp = compile(f"({_pbond}?({_patom}|({_pring})))|{_psaturate}")


def ToTokens(s: str) -> List[str]:
    """Convert AMSR string to a list of tokens

    :param s: AMSR
    :return: list of tokens
    """
    t = []
    for m in RegExp.finditer(s):
        g = m.groupdict()
        for k in ["bond", "atom", "saturate"]:
            if g[k] is not None:
                t.append(g[k])
        if g["ring"] is not None:
            t.append(g["ring"].replace(SKIP, ""))
            t.extend(SKIP * g["ring"].count(SKIP))
    return t



from rdkit import Chem
from typing import List
from re import match, escape
from .atom import Atom
from .bond import Bond
from .pibonds import PiBonds
from .bfs import BFSTree
from .parity import IsEvenParity
from .groups import DecodeGroups
from .tokens import RegExp, SKIP, L_BRACKET, R_BRACKET


def _addBond(mol, atom, i, j, bond):
    atom[i].nNeighbors += 1
    atom[j].nNeighbors += 1
    n = mol.AddBond(i, j, Chem.BondType.SINGLE)
    s = bond.rdStereo()
    if s is not None:
        mol.GetBondWithIdx(n - 1).SetStereo(s)


def _addAtom(mol, atom, a, bond, contiguous=False):
    atom.append(a)
    mol.AddAtom(a.asRDAtom())
    if not a.canBond():
        return
    j = len(atom) - 1
    for i in reversed(range(j)):
        if atom[i].canBond():
            _addBond(mol, atom, i, j, bond)
            return
    if not contiguous:
        return
    for i in reversed(range(j)):
        if atom[i].nNeighbors < atom[i].maxNeighbors:
            _addBond(mol, atom, i, j, bond)
            return


def _saturate(atom):
    for i in reversed(range(len(atom))):
        if atom[i].canBond():
            atom[i].isSaturated = True
            return


def _ring(mol, atom, ringStr, bond):
    m = match(
        f"{escape(L_BRACKET)}?([0-9]+){escape(R_BRACKET)}?({escape(SKIP)}*)", ringStr
    )
    n = int(m.group(1))
    if n < 3:
        return
    nSkip = len(m.group(2))
    for i in reversed(range(len(atom))):
        if not atom[i].canBond():
            continue
        for node in BFSTree(mol.GetAtomWithIdx(i), n - 1):
            j = node.name.GetIdx()
            if not atom[j].canBond():
                continue
            if nSkip == 0:
                _addBond(mol, atom, i, j, bond)
                return
            else:
                nSkip -= 1


def ToMol(s: str, contiguous: bool = False, useFilters: bool = True) -> Chem.Mol:
    """Convert AMSR to an RDKit Mol

    :param s: AMSR
    :param contiguous: make second pass (considering "saturated" atoms) to form bonds, rather than start new molecule
    :param useFilters: use a subset of filters from J. Chem. Inf. Model. 52, 2864 (2012) to exclude unstable or synthetically inaccessible molecules
    :return: RDKit Mol
    """
    mol = Chem.RWMol()
    atom: List[Atom] = []
    for m in RegExp.finditer(DecodeGroups(s)):
        if m.group("ring"):
            _ring(mol, atom, m.group("ring"), Bond(m.group("bond")))
        elif m.group("atom"):
            _addAtom(
                mol, atom, Atom(m.group("atom")), Bond(m.group("bond")), contiguous
            )
        elif m.group("saturate"):
            _saturate(atom)
    for a in mol.GetAtoms():
        if a.GetChiralTag() != Chem.ChiralType.CHI_UNSPECIFIED:
            n = len(a.GetBonds())
            if n < 3:
                a.SetChiralTag(Chem.ChiralType.CHI_UNSPECIFIED)
            elif not IsEvenParity([b.GetIdx() for b in a.GetNeighbors()]):
                a.InvertChirality()
    for b in mol.GetBonds():
        if b.GetStereo() in (Chem.BondStereo.STEREOZ, Chem.BondStereo.STEREOE):
            ai, aj = b.GetBeginAtom(), b.GetEndAtom()
            i, j = ai.GetIdx(), aj.GetIdx()
            ni = [c.GetIdx() for c in ai.GetNeighbors() if c.GetIdx() != j]
            nj = [c.GetIdx() for c in aj.GetNeighbors() if c.GetIdx() != i]
            if len(ni) == 0 or len(nj) == 0:
                b.SetStereo(Chem.BondStereo.STEREONONE)
            else:
                b.SetStereoAtoms(min(ni), min(nj))
    PiBonds(mol, atom, useFilters)
    for i, a in enumerate(atom):
        if a.bangs > 0 and a.canBond():
            mol.GetAtomWithIdx(i).SetNumExplicitHs(a.maxNeighbors - a.nNeighbors)
    mol = mol.GetMol()
    Chem.SanitizeMol(mol)
    Chem.AssignStereochemistry(mol)
    for a in reversed(mol.GetAtoms()):
        if atom[a.GetIdx()].canBond():
            a.SetBoolProp("_active", True)
            break

    return mol


def ToSmiles(s: str, contiguous: bool = False, useFilters: bool = True) -> str:
    """Convert AMSR to SMILES

    :param s: AMSR
    :param contiguous: make second pass (considering "saturated" atoms) to form bonds, rather than start new molecule
    :param useFilters: use a subset of filters from J. Chem. Inf. Model. 52, 2864 (2012) to exclude unstable or synthetically inaccessible molecules
    :return: SMILES
    """
    return Chem.MolToSmiles(ToMol(s, contiguous, useFilters))



__all__ = [
    "FromMol",
    "FromMolToTokens",
    "FromSmiles",
    "FromSmilesToTokens",
    "ToMol",
    "ToSmiles",
    "CheckMol",
    "CheckSmiles",
    "Groups",
    "InitializeGroups",
    "ToTokens",
    "Morph",
    "Sampler",
]

from .version import __version__
from .encode import FromMol, FromMolToTokens, FromSmiles, FromSmilesToTokens
from .decode import ToMol, ToSmiles
from .check import CheckMol, CheckSmiles
from .groups import Groups, InitializeGroups
from .tokens import ToTokens
from .morph import Morph
from .sampler import Sampler
